"""gelfista.py: FISTA like algorithm for group elastic net.

This implementation uses accelerated proximal gradient descent to solve the
optimization problem:

    min_b (1/2m)||y - b_0 - sum_j A_j@b_j||^2 +
          sum_j {sqrt{n_j}(l_1*||b_j|| + l_2*||b_j||^2)}

Here, and everywhere else, the "@" symbol is used to denote matrix
multiplication. Denoting the first part of the objective by g, and the second
part by h, the algorithm can be written as

    v = b^{k-1} + ((k-2)/(k-1))*(b^{k-1} - b^{k-2})
    b^k = prox_{t_k}(v - t_k*grad(g(v)))

Here t_k is a step length chosen by backtracking line search, and prox is the
proximal operator of h.

The functions in this module use certain variables and conventions. These will
be defined next.

The coefficients are represented using a scalar bias b_0 and a matrix B where
each row corresponds to a single group (with appropriate 0 padding). The root of
the group sizes are stored in a matrix sns broadcasted to the shape of B. The
features are stored in a 3D tensor A of size #groups x max{n_j} x m. Two helper
variables are used: a_1 = l_1*sns, and a_2 = 2*l_2*sns.
"""

import torch
import tqdm


def _prox(B, a_1, a_2, t):
    """Compute the prox operator of h.

    In this case, this reduces to a simple soft thresholding:

        prox_t(x)_0 = x_0
        prox_t(x)_j =
            if ||x_j|| > t*l_1*sqrt(n_j):
                (x_j/||x_j||)*((||x_j|| - t*l_1*sqrt(n_j)) /
                               (1 + 2*t*l_2*sqrt(n_j)))
            else:
                0

    Here x_j denotes the coefficient corresponding to the j^th group.

    Arguments have usual meanings. b_0 is not passed to the function since it is
    not changed.
    """
    # First compute assuming 'if' part of the condition.
    ta_1 = t * a_1
    norms = B.norm(p=2, dim=1, keepdim=True).expand_as(B)
    shrinkage = (norms - ta_1) / (1 + t * a_2)
    prox = B * shrinkage / norms

    # Threshold to zero if ||x_j|| is below threshold.
    thresh_cond = norms <= ta_1
    prox[thresh_cond] = 0
    return prox


def _r(A, b_0, B, y):
    """Compute y - b_0 - sum_j A_j@b_j."""
    # bmm will return a tensor with individual A_j@b_j products.
    AB = torch.bmm(A.permute(0, 2, 1), B.unsqueeze(2)).sum(dim=0).squeeze()
    return y - b_0 - AB


def _g(A, b_0, B, y, m):
    """Compute g(b)."""
    # m is the number of samples.
    r = _r(A, b_0, B, y)
    return r @ r / (2.0 * m)


def _grad(A, b_0, B, y, p, m):
    """Compute the gradient of g.

    Gradient for b_0 is 1'@r / -m, and gradient for b_j is A_j'@r / -m.
    """
    r = _r(A, b_0, B, y)
    grad_b_0 = r.sum() / -m
    # r must be broadcasted for the next operation.
    r = r.unsqueeze(0).unsqueeze(2).expand(p, m, 1)
    grad_B = torch.bmm(A, r).squeeze() / -m
    if len(grad_B.shape) < len(B.shape):
        grad_B = grad_B.unsqueeze(1)
    return grad_b_0, grad_B


def make_A(As, ns, device=torch.device("cpu"), dtype=torch.float32):
    """Create the 3D tensor A as needed by gel_solve, given a list of feature
    matrices.

    Arguments:
        As: list of feature matrices, one per group (size mxn_j).
        ns: LongTensor of group sizes.
        device: torch device (default cpu).
        dtype: torch dtype (default float32).
    """
    A = torch.zeros(
        len(ns), ns.max(), As[0].shape[0], device=device, dtype=dtype
    )
    for j, n_j in enumerate(ns):
        # Fill A[j] with A_j'.
        A[j, :n_j, :] = As[j].t()
    return A


def gel_solve(
    A,
    y,
    l_1,
    l_2,
    ns,
    b_init=None,
    t_init=None,
    ls_beta=None,
    max_iters=None,
    rel_tol=1e-6,
    verbose=False,
):
    """Solve a group elastic net problem.

    Arguments:
        A: 3D tensor of features as described in the header, returned by
            make_A.
        y: tensor vector of predictions.
        l_1: the 2-norm coefficient.
        l_2: the squared 2-norm coefficient.
        ns: iterable of group sizes.
        b_init: tuple (b_0, B) to initialize b.
        t_init: initial step size to start line search; if None, it's set to
            the value from the previous iteration.
        ls_beta: shrinkage factor for line search; if None, no line search is
            performed, and t_init is used as the step size for every iteration.
        max_iters: maximum number of iterations; if None, there is no limit.
        rel_tol: tolerance for exit criterion.
        verbose: boolean to enable/disable verbosity.
    """
    p = len(ns)
    m = len(y)
    device = A.device
    dtype = A.dtype
    y = y.to(device, dtype)

    # Create initial values if not specified.
    if b_init is None:
        b_init = 0.0, torch.zeros(p, max(ns), device=device, dtype=dtype)

    b_0, B = b_init
    b_0_prev, B_prev = b_0, B
    if not isinstance(ns, torch.Tensor):
        ns = torch.tensor(ns)
    sns = ns.to(device, dtype).sqrt().unsqueeze(1).expand_as(B)
    a_1 = l_1 * sns
    a_2 = 2 * l_2 * sns
    k = 1  # iteration number
    t = 1  # initial step length (used if t_init is None)
    pbar_stats = {}  # stats for the progress bar
    pbar = tqdm.tqdm(
        desc="Solving gel with FISTA (l_1 {:.2g}, l_2 {:.2g})".format(l_1, l_2),
        disable=not verbose,
    )

    while True:
        # Compute the v terms.
        mom = (k - 2) / (k + 1.0)
        v_0 = b_0 + mom * (b_0 - b_0_prev)
        V = B + mom * (B - B_prev)
        g_v = _g(A, v_0, V, y, m)
        grad_v_0, grad_V = _grad(A, v_0, V, y, p, m)

        b_0_prev, B_prev = b_0, B

        # Adjust the step size with backtracking line search.
        if t_init is not None:
            t = t_init
        while True:
            # Compute the update based on gradient, then apply prox.
            b_0 = v_0 - t * grad_v_0
            B = V - t * grad_V
            B = _prox(B, a_1, a_2, t)

            if ls_beta is None:
                # Don't perform line search.
                break

            # The line search condition is to exit when g(b) <= g(v) +
            # grad_v'@(b - v) + (1/2t)||b - v||^2.
            g_b = _g(A, b_0, B, y, m)
            b0_v0_diff = b_0 - v_0
            B_V_diff = B - V
            # grad_v'@(b - v):
            c_2 = grad_v_0 * b0_v0_diff + (grad_V * B_V_diff).sum()
            # (1/2t)||b - v||^2:
            c_3 = (b0_v0_diff ** 2 + (B_V_diff ** 2).sum()) / (2.0 * t)

            if g_b <= g_v + c_2 + c_3:
                break
            else:
                t *= ls_beta

        # Compute relative change in b.
        b_0_diff = b_0 - b_0_prev
        B_diff = B - B_prev
        delta_norm = (b_0_diff ** 2 + (B_diff ** 2).sum()).sqrt()
        b_norm = (b_0 ** 2 + (B ** 2).sum()).sqrt()

        pbar_stats["t"] = "{:.2g}".format(t)
        pbar_stats["rel change"] = "{:.2g}".format((delta_norm / b_norm).item())
        pbar.set_postfix(pbar_stats)
        pbar.update()

        # Check max iterations exit criterion.
        if max_iters is not None and k == max_iters:
            break
        k += 1

        # Check tolerance exit criterion. Break if the relative change in 2-norm
        # between b and b_prev is less than tol.
        # k > 2 ensures that at least 2 iterations are completed.
        if delta_norm.item() <= rel_tol * b_norm.item() and k > 2:
            break

    pbar.close()
    return b_0.item(), B



"""ridgepaths.py: solution paths for ridge regression."""

import torch
import tqdm


def ridge_paths(X, y, support, lambdas, summ_fun, verbose=False):
    """Solve ridge ridgression for a sequence of regularization values,
    and return the summary for each.

    The multiple solutions are obtained efficiently using the Woodbury identity.
    With X (p x m) representing the feature matrix, and y (m x 1) the outcomes,
    the ridge solution is given by

        b = (X@X' + l*I)^{-1}@X@y

    where l is the regularization coefficient. This can be reduced to

        (1/l)*(X@y - X@V(e + l*I)^{-1}@(X@V)'@X@y)

    where V@e@V' is the eigendecomposition of X'@X. Since (e + l*I) is a
    diagonal matrix, its inverse can be performed efficiently simply by taking
    the reciprocal of the diagonal elements. Then, (X@V)'@X@y is a vector; so it
    can be multiplied by (e + l*I)^{-1} just by scalar multiplication.

    Arguments:
        X: pxm tensor of features (where m is the number of samples);
            X should be centered (each row should have mean 0).
        y: tensor (length m vector) of outcomes.
        support: LongTensor vector of features to use. This vector should
            contain indices. It can also be None to indicate empty support.
            X should already be indexed using support. This argument is simply
            passed to the summary function.
        lambdas: list of regularization values for which to solve the problem.
        summ_fun: a function that takes (support, b) and returns an arbitrary
            summary.
        verbose: enable/disable the progress bar.

    The function returns a dictionary mapping lambda values to their summaries.
    """
    summaries = {}

    if support is None:
        # Nothing to do.
        for l in tqdm.tqdm(
            lambdas, desc="Solving ridge regressions", disable=not verbose
        ):
            summaries[l] = summ_fun(None, None)
        return summaries

    # Setup.
    _, S, V = torch.svd(X)
    e = S ** 2
    p = X @ y
    Q = X @ V
    r = Q.t() @ p  # (X@V)'@X@y

    # Main loop.
    for l in tqdm.tqdm(
        lambdas, desc="Solving ridge regressions", disable=not verbose
    ):
        b = (1.0 / l) * (p - Q @ (r / (e + l)))
        summaries[l] = summ_fun(support, b)
    return summaries



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['gel']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.10,<2.0', 'tqdm>=4.0,<5.0']

extras_require = \
{'test': ['scipy', 'cvxpy>=1.0,<2.0', 'cvxopt']}

setup_kwargs = {
    'name': 'torchgel',
    'version': '1.0.0',
    'description': 'PyTorch implementation of group elastic net',
    'long_description': "# torch-gel\n\nThis package provides PyTorch implementations to solve the group elastic net\nproblem. Let _A<sub>j</sub>_ (_j = 1 … p_) be feature matrices of sizes _m ×\nn<sub>j</sub>_ (_m_ is the number of samples, and _n<sub>j</sub>_ is the number\nof features in the _j_<sup>th</sup> group), and let _y_ be an _m × 1_ vector of\nthe responses. Group elastic net finds coefficients _β<sub>j</sub>_, and a bias\n_β<sub>0</sub>_ that solve the optimization problem\n\n> min _<sub>β<sub>0</sub>, …, β<sub>p</sub></sub>_\n>     _½ ║y - β<sub>0</sub> - ∑ A<sub>j</sub> β<sub>j</sub>║<sup>2</sup>_\n>     + _m ∑ √n<sub>j</sub> (λ<sub>1</sub>║β<sub>j</sub>║_\n>                           _+ λ<sub>2</sub>║β<sub>j</sub>║<sup>2</sup>)._\n\nHere _λ<sub>1</sub>_ and _λ<sub>2</sub>_ are scalar coefficients that control\nthe amount of 2-norm and squared 2-norm regularization. This 2-norm\nregularization encourages sparsity at the group level; entire _β<sub>j</sub>_\nmight become 0. The squared 2-norm regularization is in similar spirit to\nelastic net, and addresses some of the issues of lasso. Note that group elastic\nnet includes as special cases group lasso (_λ<sub>2</sub> = 0_), ridge\nregression (_λ<sub>1</sub> = 0_), elastic net (each _n<sub>j</sub> = 1_), and\nlasso (each _n<sub>j</sub> = 1_ and _λ<sub>2</sub> = 0_). The optimization\nproblem is convex, and can be solved efficiently. This package provides two\nimplementations; one based on proximal gradient descent, and one based on\ncoordinate descent.\n\n## Installation\nInstall with `pip`\n\n```bash\npip install torchgel\n```\n\n`tqdm` (for progress bars), and numpy are pulled in as dependencies. PyTorch\n(`v1.0+`) is also needed, and needs to be installed manually. Refer to the\n[PyTorch website](<http://pytorch.org>) for instructions.\n\n## Usage\n[`examples/main.ipynb`](examples/main.ipynb) is a Jupyter notebook that walks\nthrough using the package for a typical use-case. A more formal description of\nthe functions follows; and for details about the algorithms, refer to the\ndocstrings of files in the `gel` directory.\n\n### Solving Single Instances\nThe modules `gel.gelfista` and `gel.gelcd` provide implementations based on\nproximal gradient descent and coordinate descent respectively. Both have similar\ninterfaces, and expose two main public functions: `make_A` and `gel_solve`. The\nfeature matrices should be stored in a list (say `As`) as PyTorch tensor\nmatrices, and the responses should be stored in a PyTorch vector (say `y`).\nAdditionally, the sizes of the groups (_n<sub>j</sub>_) should be stored in a\nvector (say `ns`). First use the `make_A` function to convert the feature\nmatrices into a suitable format:\n\n```python\nA = make_A(As, ns)\n```\n\nThen pass `A`, `y` and other required arguments to `gel_solve`. The general\ninterface is::\n\n```python\nb_0, B = gel_solve(A, y, l_1, l_2, ns, **kwargs)\n```\n\n`l_1` and `l_2` are floats representing _λ<sub>1</sub>_ and _λ<sub>2</sub>_\nrespectively. The method returns a float `b_0` representing the bias and a\nPyTorch matrix `B` holding the other coefficients. `B` has size _p ×_\nmax<sub>_j_</sub> _n<sub>j</sub>_ with suitable zero padding. The following\nsections cover additional details for the specific implementations.\n\n#### Proximal Gradient Descent (FISTA)\nThe `gel.gelfista` module contains a proximal gradient descent implementation.\nIt's usage is just as described in the template above. Refer to the docstring\nfor `gel.gelfista.gel_solve` for details about the other arguments.\n\n#### Coordinate Descent\nThe `gel.gelcd` module contains a coordinate descent implementation. Its usage\nis a bit more involved than the FISTA implementation. Coordinate descent\niteratively solves single blocks (each corresponding to a single\n_β<sub>j</sub>_). There are multiple solvers provided to solve the individual\nblocks. These are the `gel.gelcd.block_solve_*` functions. Refer to their\ndocstrings for details about their arguments. `gel.gelcd.gel_solve` requires\npassing a block solve function and its arguments (as a dictionary). Refer to\nits docstring for further details.\n\n### Solution Paths\n`gel.gelpaths` provides a wrapper function `gel_paths` to solve the group\nelastic net problem for multiple values of the regularization coefficients. It\nimplements a two-stage process. For a given _λ<sub>1</sub>_ and _λ<sub>2</sub>_,\nfirst the group elastic net problem is solved and the feature blocks with\nnon-zero coefficients is extracted (the support). Then ridge regression models\nare learned for each of several provided regularization values. The final model\nis summarized using an arbitrary provided summary function, and the summary for\neach combination of the regularization values is returned as a dictionary. The\ndocstring contains more details. `gel.ridgepaths` contains another useful function,\n`ridge_paths` which can efficiently solve ridge regression for multiple\nregularization values.\n\n## Citation\nIf you find this code useful in your research, please cite\n\n```\n@misc{koushik2017torchgel,\n  author = {Koushik, Jayanth},\n  title = {torch-gel},\n  year = {2017},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/jayanthkoushik/torch-gel}},\n}\n```\n",
    'author': 'Jayanth Koushik',
    'author_email': 'jnkoushik@gmail.com',
    'url': 'https://github.com/jayanthkoushik/torch-gel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)
